#define INCL_DOSFILEMGR          /* File Manager values */
#define INCL_DOSERRORS           /* DOS Error values    */
#include <os2.h>
#include <gl.h>
#include <glu.h>
#include <aux.h>
#include <math.h>
#include <stdlib.h>
#include <ithread.hpp>

#include "glwindow.hpp"

#define FOVY            40.0
#define MINVIEWDISTANCE 100.0
#define MAXVIEWDISTANCE 2000.0

// GLpaint destructor handler
Boolean GLPaintDestroy::dispatchHandlerEvent( IEvent &event )
{
   Boolean stopDispatching = false;
   switch( event.eventId() )
   {
      case WM_DESTROY:
         stopDispatching = destroy( event );
         break;
   }
   return( stopDispatching );
}

Boolean GLPaintDestroy::destroy( IEvent &event )
{
   // stop the timer
   paintHandler->timer.stop();
   return false;
}

// constructors and destructors
GLPaintHandler::GLPaintHandler( IWindow &_window )
: window( _window ), paintDestroy( this ), rotationVal( 0.0 )
{
   // activate the destructor handler
   paintDestroy.handleEventsFor( &_window );

   // set timer
   timer
      .setInterval( 100 )
      .start( new ITimerMemberFn0<GLPaintHandler>( *this, GLPaintHandler::timerFn ));
}

GLPaintHandler::~GLPaintHandler()
{
   HAB hab = IThread::current().anchorBlock();

   // unbind the current context
   pglMakeCurrent( hab, NULL, NULL );

   // destroy the context we created
   pglDestroyContext( hab, hgc );
}

void GLPaintHandler::initOpenGL()
{
   int attribList[] = { PGL_RGBA, PGL_RED_SIZE, 2, PGL_GREEN_SIZE, 2,
      PGL_BLUE_SIZE, 2, PGL_ALPHA_SIZE, 1, PGL_DOUBLEBUFFER, 0 };

   // get a handle to anchor block
   HAB hab =  IThread::current().anchorBlock();

   // create a visual config
   PVISUALCONFIG pVisualConfig = pglChooseConfig( hab, attribList );

   // create a direct OpenGL context
   hgc = pglCreateContext( hab, pVisualConfig, NULL, true );

   // and make the context we got the current one
   Boolean val = pglMakeCurrent( hab, hgc, window.handle() );

   // enable depth buffer
   glEnable( GL_DEPTH_TEST );

   glClearColor( 0.0, 0.0, 0.0, 1.0 );
   glDrawBuffer( GL_FRONT_AND_BACK );
   glClear( GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT  );
   glDrawBuffer( GL_BACK );

   glPolygonMode( GL_FRONT_AND_BACK, GL_FILL );
   glShadeModel( GL_FLAT );

   // free the visual config created above
   delete [] pVisualConfig;
}

// openGL settings
void GLPaintHandler::setWindowSize( ISize size )
{
   // setup the 3D perspective window
   glMatrixMode( GL_PROJECTION );
   glLoadIdentity();
   gluPerspective( FOVY, (float)window.size().width() / (float)window.size().height(),
                  MINVIEWDISTANCE, MAXVIEWDISTANCE );
   glMatrixMode( GL_MODELVIEW );
   // make a viewport the acual size of the window
   glViewport( 0, 0, size.width(), size.height() );
}

#define STEP_SIZE 10
Boolean GLPaintHandler::paintWindow( IPaintEvent &event )
{
   // clear the buffer
   glClear( GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT );

   // no rotations
   glLoadIdentity();

   // move the image back from the camera
   glTranslatef( 0.0, 0.0, -1000.0 );
   glRotatef( rotationVal, 0.0, 1.0, 0.0 );
   glRotatef( 20.0, 1.0, 0.0, 0.0 );

   // draw the object
   for( int i=-200; i<200; i+=STEP_SIZE )
   {
      for( int j=-200; j<200; j+=STEP_SIZE )
      {
         // draw the polygon
         glColor3f( 0.0, 0.0, 1.0 );
         glBegin( GL_TRIANGLES );
            glVertex3f( i, j, 0 );
            glVertex3f( i, j+STEP_SIZE, 0 );
            glVertex3f( i+STEP_SIZE, j+STEP_SIZE, 0 );

            glVertex3f( i, j, 0 );
            glVertex3f( i+STEP_SIZE, j+STEP_SIZE, 0 );
            glVertex3f( i+STEP_SIZE, j, 0 );
         glEnd();
      }
   }
   // swap the back draw buffer to the front
   pglSwapBuffers( IThread::current().anchorBlock(), window.handle() );

   // return true to indicate no additional processing needed
   return( true );
}

void GLPaintHandler::timerFn()
{
   // recalculate positions
   rotationVal += 5.0;
   if( rotationVal > 360.0 )
      rotationVal -= 360.0;

   window.refresh();
}

// window resize handler
Boolean GLResizeHandler::windowResize(IResizeEvent &event )
{
   // tell the paint handler the new size
   window->glPaintHandler().setWindowSize( event.newSize() );
   return false;
}

// constructor
GLWindow::GLWindow()
: IFrameWindow( "OpenGL Demonstration 10", 0x1000, IFrameWindow::defaultStyle()
         | IFrameWindow::noMoveWithOwner | IWindow::synchPaint ),
   canvas( 0x8008, this, this ),
   paintHandler( canvas ),
   resizeHandler( this )
{
   // setup
   setClient( &canvas );
   setDestroyOnClose( true );

   // draw on screen
   sizeTo( ISize( 400, 400 ));
   setFocus();
   show();

   // initialize openGL
   paintHandler.initOpenGL();

   // set the size of the 3D window
   paintHandler.setWindowSize( canvas.size() );

   // attach the handlers
   paintHandler.handleEventsFor( &canvas );
   resizeHandler.handleEventsFor( &canvas );
}

// destructor
GLWindow::~GLWindow()
{
   // detach the handlers
   paintHandler.stopHandlingEventsFor( &canvas );
   resizeHandler.stopHandlingEventsFor( &canvas );
}
